from urllib.parse import urlparse, parse_qsl, urlunparse, urlencode, parse_qs
import pprint

from .excs import HTTPError, AuthError


DEFAULT_UPLOAD_CHUNK_SIZE = 1024 * 1024


class ResumableUpload:
    """
    Resumable Upload Object. Works in conjuction with media upload 
    
    Arguments:

        file_path (str): Full path of the file to be uploaded
        
        upload_path (str): The URI path to be used for upload. Should be used in conjunction with the rootURL property at the API-level.
        
        multipart (bool): True if this endpoint supports upload multipart media.

        chunksize (int): Size of a chunk of bytes that a session should read at a time when uploading in multipart.

    """

    def __init__(self, file_path, multipart=None, chunk_size=None, upload_path=None):
        self.file_path = file_path
        self.upload_path = upload_path
        self.multipart = multipart
        if chunk_size is None:
            chunk_size = DEFAULT_UPLOAD_CHUNK_SIZE
        self.chunk_size = chunk_size


class MediaUpload:
    """

    Media Upload

    Arguments:

        file_path (str): Full path of the file to be uploaded
        
        upload_path (str): The URI path to be used for upload. Should be used in conjunction with the rootURL property at the API-level.
        
        mime_range (list): list of MIME Media Ranges for acceptable media uploads to this method.
        
        max_size (int): Maximum size of a media upload in bytes
        
        multipart (bool): True if this endpoint supports upload multipart media.

        chunksize (int): Size of a chunk of bytes that a session should read at a time when uploading in multipart.
        
        resumable (aiogoogle.models.ResumableUplaod): A ResumableUpload object

        validate (bool): Whether or not a session should validate the upload size before sending

    """

    def __init__(
        self,
        file_path,
        upload_path=None,
        mime_range=None,
        max_size=None,
        multipart=False,
        chunk_size=None,
        resumable=None,
        validate=True,
    ):
        self.file_path = file_path
        self.upload_path = upload_path
        self.mime_range = mime_range
        self.max_size = max_size
        self.multipart = multipart
        if chunk_size is None:
            chunk_size = DEFAULT_UPLOAD_CHUNK_SIZE
        self.chunk_size = chunk_size
        self.resumable = resumable
        self.validate = validate


class MediaDownload:
    """
    Media Download

    Arguments:

        file_path (str): Full path of the file to be downloaded

    """

    def __init__(self, file_path):
        self.file_path = file_path


class Request:
    """
    Request class for the whole library. Auth Managers, GoogleAPI and Sessions should all use this.

    .. note::
        
        For HTTP body, only pass one of the following params:
            
            - json: json as a dict
            - data: www-url-form-encoded form as a dict/ bytes/ text/ 


    Parameters:

        method (str): HTTP method as a string (upper case) e.g. 'GET'
        
        url (str): full url as a string. e.g. 'https://example.com/api/v1/resource?filter=filter#something

        batch_url (str): full url of for sending this request in a batch
        
        json (dict): json as a dict
        
        data (any): www-url-form-encoded form as a dict/ bytes/ text/ 
        
        headers (dict): headers as a dict
        
        media_download (aiogoogle.models.MediaDownload): MediaDownload object
        
        media_upload (aiogoogle.models.MediaUpload): MediaUpload object
        
        timeout (int): Individual timeout for this request

        callback (callable): Synchronous callback that takes the content of the response as the only argument. Should also return content.
        """

    def __init__(
        self,
        method=None,
        url=None,
        batch_url=None,
        headers=None,
        json=None,
        data=None,
        media_upload=None,
        media_download=None,
        timeout=None,
        callback=None,
    ):
        self.method = method
        self.url = url
        self.batch_url = batch_url
        if headers is None:
            self.headers = {}
        else:
            self.headers = headers
        self.data = data
        self.json = json
        self.media_upload = media_upload
        self.media_download = media_download
        self.timeout = timeout
        self.callback = callback

    def _add_query_param(self, query: dict):
        if not self.url:
            raise TypeError("no url to add query to")

        url = self.url
        if "?" not in url:
            if url.endswith("/"):
                url = url[:-1]
            url += "?"
        else:
            url += "&"
        query = urlencode(query)
        url += query
        self.url = url

    def _rm_query_param(self, name: str):
        u = urlparse(self.url)
        query = parse_qs(u.query)
        query.pop(name, None)
        u = u._replace(query=urlencode(query, True))
        self.url = urlunparse(u)

    @classmethod
    def batch_requests(cls, *requests):
        """
        Given many requests, will create a batch request per https://developers.google.com/discovery/v1/batch

        Arguments:

            *requests (aiogoogle.models.Request): Request objects

        Returns:

            aiogoogle.models.Request:
        """
        raise NotImplementedError

    @classmethod
    def from_response(cls, response):
        return Request(
            url=response.url,
            headers=response.headers,
            json=response.json,
            data=response.data,
        )


class Response:
    """
    Respnse Object

    Arguments:

        status_code (int): HTTP Status code

        headers (dict): HTTP response headers

        url (str): Request URL

        json (dict): Json Response if any

        data (any): data

        reason (str): reason for http error if any

        req (aiogoogle.models.Request): request that caused this response

        download_file (str): path of the download file specified in the request

        upload_file (str): path of the upload file specified in the request

        session_factory (aiogoogle.sessions.abc.AbstractSession): A callable implementation of aiogoogle's session interface

    Attributes:

        content (any): equals either ``self.json`` or ``self.data``
    """

    def __init__(
        self,
        status_code=None,
        headers=None,
        url=None,
        json=None,
        data=None,
        reason=None,
        req=None,
        download_file=None,
        upload_file=None,
        session_factory=None,
    ):
        if json and data:
            raise TypeError("Pass either json or data, not both.")

        self.status_code = status_code
        self.headers = headers
        self.url = url
        self.json = json
        self.data = data
        self.reason = reason
        self.req = req
        self.download_file = download_file
        self.upload_file = upload_file
        self.session_factory = session_factory

    @staticmethod
    async def _next_page_generator(
        prev_res,
        session_factory,
        req_token_name=None,
        res_token_name=None,
        json_req=False,
    ):
        prev_url = None
        while prev_res is not None:

            # Avoid infinite looping if google sent the same token twice
            if prev_url == prev_res.req.url:
                break
            prev_url = prev_res.req.url

            # yield
            yield prev_res.content

            # get request for next page
            next_req = prev_res.next_page(
                req_token_name=req_token_name,
                res_token_name=res_token_name,
                json_req=json_req,
            )
            if next_req is not None:
                async with session_factory() as sess:
                    prev_res = await sess.send(next_req, full_res=True)
            else:
                prev_res = None

    def __call__(
        self,
        session_factory=None,
        req_token_name=None,
        res_token_name=None,
        json_req=False,
    ):
        """
        Returns a generator that yields the contents of the next pages if any (and this page as well)

        Arguments:

            session_factory (aiogoogle.sessions.abc.AbstractSession): A session factory

            req_token_name (str):
            
                * name of the next_page token in the request

                * Default: "pageToken"
            
            res_token_name (str): 
            
                * name of the next_page token in json response

                * Default: "nextPageToken"

            json_req (dict): Normally, nextPageTokens should be sent in URL query params. If you want it in A json body, set this to True

        Returns:

            async generator: self._next_page_generator (staticmethod)
        """
        if session_factory is None:
            session_factory = self.session_factory
        return self._next_page_generator(
            self, session_factory, req_token_name, res_token_name, json_req
        )

    def __aiter__(self):
        return self._next_page_generator(self, self.session_factory)

    def __iter__(self):
        raise TypeError(
            'You probably forgot to use an "async for" statement instead of just a "for" statement.'
        )

    @property
    def content(self):
        return self.json or self.data

    def next_page(
        self, req_token_name=None, res_token_name=None, json_req=False
    ) -> Request:
        """
        Method that returns a request object that requests the next page of a resource

        Arguments:

            req_token_name (str):
            
                * name of the next_page token in the request

                * Default: "pageToken"
            
            res_token_name (str): 
            
                * name of the next_page token in json response

                * Default: "nextPageToken"

            json_req (dict): Normally, nextPageTokens should be sent in URL query params. If you want it in A json body, set this to True

        Returns:

            A request object (aiogoogle.models.Request):
        """
        if req_token_name is None:
            req_token_name = "pageToken"
        if res_token_name is None:
            res_token_name = "nextPageToken"
        res_token = self.json.get(res_token_name, None)
        if res_token == "":
            res_token = None
        if res_token is None:
            return None
        # request = Request.from_response(self)
        request = self.req
        if json_req:
            request.json[req_token_name] = res_token
        else:
            request._rm_query_param(req_token_name)
            request._add_query_param({req_token_name: res_token})
        return request

    @property
    def error_msg(self):
        if self.json is not None and self.json.get("error") is not None:
            return pprint.pformat(self.json["error"])

    def raise_for_status(self):
        if self.status_code >= 400:
            if self.error_msg is not None:
                self.reason = "\n\n" + self.reason + "\n\nContent:\n" + self.error_msg
            self.reason = "\n\n" + self.reason + "\n\nRequest URL:\n" + self.req.url
            if self.status_code == 401:
                raise AuthError(msg=self.reason, req=self.req, res=self)
            else:
                raise HTTPError(msg=self.reason, req=self.req, res=self)

    def __str__(self):
        return str(self.content)

    def __repr__(self):
        return f"Aiogoogle response model. Status: {str(self.status_code)}"



__all__ = []


def _safe_getitem(dct, *keys):
    for key in keys:
        try:
            dct = dct[key]
        except (KeyError):
            return None
    return dct


class _dict(dict):  # pragma: no cover
    """ A simple dict subclass for use with Creds modelling. No surprises """

    def __init__(self, *args, **kwargs):  # pragma: no cover
        super(_dict, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):  # pragma: no cover
        return self.get(attr)

    def __setattr__(self, key, value):  # pragma: no cover
        self.__setitem__(key, value)

    def __setitem__(self, key, value):  # pragma: no cover
        super(_dict, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):  # pragma: no cover
        self.__delitem__(item)

    def __delitem__(self, key):  # pragma: no cover
        super(_dict, self).__delitem__(key)
        del self.__dict__[key]



__all__ = ["AiohttpSession"]

import asyncio
from json import JSONDecodeError

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientResponseError, ContentTypeError
import aiofiles
from aiofiles import os as async_os
import async_timeout

from ..models import Response
from .abc import AbstractSession
from ..excs import HTTPError, ValidationError
from .common import _call_callback


async def _get_file_size(full_file_path):
    stat = await async_os.stat(full_file_path)
    return stat.st_size


async def _aiter_file(file_name, chunk_size):
    """ Async file generator """
    async with aiofiles.open(file_name, "rb") as f:
        chunk = await f.read(chunk_size)
        while chunk:
            yield chunk
            chunk = await f.read(chunk_size)


class AiohttpSession(ClientSession, AbstractSession):
    async def send(
        self,
        *requests,
        timeout=None,
        full_res=False,
        raise_for_status=True,
        session_factory=None,
    ):
        async def resolve_response(request, response):
            data = None
            json = None
            download_file = None
            upload_file = None

            # If downloading file:
            if request.media_download:
                download_file = request.media_download.file_path
                async with aiofiles.open(download_file, "wb+") as f:
                    async for line in response.content:
                        await f.write(line)
            else:
                if response.status != 204:  # If no (no content)
                    try:
                        json = await response.json()
                    except (JSONDecodeError, ContentTypeError):
                        try:
                            data = await response.text()
                        except ContentTypeError:
                            try:
                                data = await response.read()
                            except ContentTypeError:
                                data = None

            if request.media_upload:
                upload_file = request.media_upload.file_path

            return Response(
                url=str(response.url),
                headers=response.headers,
                status_code=response.status,
                json=json,
                data=data,
                reason=response.reason if getattr(response, "reason") else None,
                req=request,
                download_file=download_file,
                upload_file=upload_file,
                session_factory=session_factory,
            )

        async def fire_request(request):
            request.headers["Accept-Encoding"] = "gzip"
            request.headers["User-Agent"] = "Aiogoogle Aiohttp (gzip)"
            if request.media_upload:
                # Validate
                if (
                    request.media_upload.validate is True
                    and request.media_upload.max_size is not None
                ):
                    size = await _get_file_size(request.media_upload.file_path)
                    max_size = request.media_upload.max_size
                    if size > max_size:
                        raise ValidationError(
                            f'"{request.media_upload.file_path}" has a size of {size/1000}KB. Max upload size for this endpoint is: {max_size/1000}KB.'
                        )

                # If multipart pass a file async generator
                if request.media_upload.multipart is True:
                    return await self.request(
                        method=request.method,
                        url=request.media_upload.upload_path,
                        headers=request.headers,
                        data=_aiter_file(
                            request.media_upload.file_path,
                            request.media_upload.chunk_size,
                        ),
                        json=request.json,
                        timeout=request.timeout,
                    )
                # Else load file to memory and send
                else:
                    async with aiofiles.open(
                        request.media_upload.file_path, "rb"
                    ) as file:
                        read_file = await file.read()
                        return await self.request(
                            method=request.method,
                            url=request.media_upload.upload_path,
                            headers=request.headers,
                            data=read_file,
                            json=request.json,
                            timeout=request.timeout,
                        )
            # Else, if no file upload
            else:
                return await self.request(
                    method=request.method,
                    url=request.url,
                    headers=request.headers,
                    data=request.data,
                    json=request.json,
                    timeout=request.timeout,
                )

        # ----------------- send sequence ------------------#
        async def get_response(request):
            response = await fire_request(request)
            response = await resolve_response(request, response)
            if raise_for_status is True:
                response.raise_for_status()
            response = _call_callback(request, response)
            return response

        async def get_content(request):
            response = await get_response(request)
            return response.content

        # ----------------- /send sequence ------------------#

        async def schedule_tasks():
            if full_res is True:
                tasks = [
                    asyncio.create_task(get_response(request)) for request in requests
                ]
            else:
                tasks = [
                    asyncio.create_task(get_content(request)) for request in requests
                ]
            return await asyncio.gather(*tasks, return_exceptions=False)

        session_factory = self.__class__ if session_factory is None else None

        if timeout is not None:
            async with async_timeout.timeout(timeout):
                results = await schedule_tasks()
        else:
            results = await schedule_tasks()

        return (
            results[0] if isinstance(results, list) and len(results) == 1 else results
        )



__all__ = ["GoogleAPI", "Resource", "Method"]

import re
import warnings
from urllib.parse import urlencode
from functools import wraps
import datetime

from .excs import ValidationError
from .utils import _dict, _safe_getitem
from .excs import ValidationError
from .models import MediaDownload, MediaUpload, ResumableUpload, Request
from .validate import validate as validate__


# These are the hard-coded kwargs in Method.__call__
# They're used for testing whether those names will collide with any of the url parameters that are provided by any of the discovery docs.
# If collisions were to be found, that would mean that the user won't be able to pass a url_parameter that shares the same name with any of the RESERVED_KEYWORDS.
RESERVED_KEYWORDS = [
    "validate",
    "data",
    "json",
    "upload_file",
    "download_file",
    "timeout",
]

# From: https://github.com/googleapis/google-api-python-client/blob/master/googleapiclient/discovery.py
# Parameters accepted by the stack, but not visible via discovery.
STACK_QUERY_PARAMETERS = frozenset(["trace", "pp", "strict"])
STACK_QUERY_PARAMETER_DEFAULT_VALUE = {"type": "string", "location": "query"}

# TODO: etagRequired: {
#    type: "boolean",
#    description: "Whether this method requires an ETag to be specified. The ETag is sent as an HTTP If-Match or If-None-Match header."
#    }
# Note: etagRequired is only mentioned once in all of the discovery documents available from Google. (In discovery_service-v1. So, it isn't actually being used)


def _toggle2x_dashed_params(f):
    @wraps(f)
    def wrapper(
        self,
        validate=None,
        data=None,
        json=None,
        upload_file=None,
        download_file=None,
        timeout=None,
        **uri_params,
    ):
        """
        Momentarily adds back '-' to url parameters and passed uri_params
        in order to be processed correctly and comply with the disc doc
        Reverts back to '_' after wrapped function is done
        """
        # unfix urls
        uri_params = self._add_dash_user_uri_params(uri_params)

        # unfix params
        self._method_specs["parameters"] = self._add_dash_params(
            self._method_specs.get("parameters")
        )
        self._global_parameters = self._add_dash_params(self._global_parameters)

        # Run function
        results = f(
            self,
            validate,
            data,
            json,
            upload_file,
            download_file,
            timeout,
            **uri_params,
        )

        # fix params again
        self._method_specs["parameters"] = self._rm_dash_params(
            self._method_specs.get("parameters")
        )
        self._global_parameters = self._rm_dash_params(self._global_parameters)

        return results

    return wrapper


class Method:
    def __init__(
        self,
        name,
        method_specs,
        global_parameters,
        schemas,
        root_url,
        service_path,
        batch_path,
        validate,
    ):
        # Replaces '-'s with '_'s and preserve old names to revert back to them after this method is called
        global_parameters = self._rm_dash_params(global_parameters)
        method_specs["parameters"] = self._rm_dash_params(
            method_specs.get("parameters")
        )

        self.name = name
        self._method_specs = method_specs
        self._global_parameters = global_parameters
        self._schemas = schemas

        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path

        # Not gonna use it, because it's useless. Not sure though.
        if (
            self["useMediaDownloadService"] is True
            and self["supportsMediaDownload"] is True
        ):
            self._download_base_url = self._root_url + "download/" + self._service_path
        else:
            self._download_base_url = None

        self._base_url = self._root_url + self._service_path
        self._batch_url = self._root_url + self._batch_path

        self._should_validate = validate

    # ---- Changes URL parameters with a "-" to "_" -----#
    # Depends on how you view it, but this section also changes robots with small mouths to robots with big mouths

    @staticmethod
    def _rm_dash_params(param_set) -> dict:
        if param_set:
            for name, schema in param_set.items():
                if "-" in name:
                    new_name = name.replace("-", "_")  # See?!
                    schema["orig_name"] = name
                    param_set[new_name] = schema
                    del param_set[name]
        return param_set

    @staticmethod
    def _add_dash_params(param_set) -> dict:
        if param_set:
            # list() forces the creation of a new copy in memory
            # to avoid having the dict size changing during iteration
            for name, schema in list(param_set.items()):
                if "orig_name" in schema:
                    param_set[schema["orig_name"]] = schema
                    del param_set[name]
        return param_set

    def _add_dash_user_uri_params(self, uri_params):
        for k, v in uri_params.items():
            if "_" in k:
                if k in self.parameters:
                    if "orig_name" in self.parameters[k]:
                        uri_params[self.parameters[k]["orig_name"]] = v
                        del uri_params[k]
        return uri_params

    # ---- / Changes URL parameters with a "-" to "_" -----#

    @property
    def request(self) -> dict:
        """ Returns expected request body """
        body = self["request"]
        if body.get("$ref"):
            return self._schemas.get(body["$ref"])
        else:
            return body

    @property
    def response(self) -> dict:
        """ Retruns expected response body """
        body = self["response"]
        if body.get("$ref"):
            return self._schemas.get(body["$ref"])
        else:
            return body

    @property
    def parameters(self) -> dict:
        """ 
        Parameters property

        Returns:

            dict: All parameters that this method can take as described in the discovery document
        """
        if not self._global_parameters and not self["parameters"]:
            return {}
        elif not self._global_parameters:
            return self["parameters"]
        elif not self["parameters"]:
            return self._global_parameters
        else:
            return {**self["parameters"], **self._global_parameters}

    @property
    def optional_parameters(self) -> [str, str]:
        """
        Optional Parameters

        Returns:

            list: List of the names of optional parameters this method takes
        """
        return (
            [
                parameter_name
                for parameter_name, parameter_info in self.parameters.items()
                if parameter_info.get("required") is not True
            ]
            if self.parameters
            else []
        )

    @property
    def required_parameters(self) -> [str, str]:
        """
        Required Parameters

        Returns:

            list: List of the names of required parameters this method takes
        """
        return (
            [
                parameter_name
                for parameter_name, parameter_info in self.parameters.items()
                if parameter_info.get("required") is True
            ]
            if self.parameters
            else []
        )

    @property
    def path_parameters(self) -> [str, str]:
        """
        Path Parameters

        Returns:

            list: List of the names of path parameters this method takes
        """
        return (
            [
                param_name
                for param_name, param_info in self.parameters.items()
                if param_info.get("location") == "path"
            ]
            if self.parameters
            else []
        )

    @property
    def query_parameters(self) -> [str, str]:
        """
        Query Parameters

        Returns:

            list: List of the names of Query parameters this method takes
        """
        return (
            [
                param_name
                for param_name, param_info in self.parameters.items()
                if param_info.get("location") == "query"
            ]
            if self.parameters
            else []
        )

    @property
    def required_query_parameters(self) -> [str, str]:
        """
        Required Query Parameters

        Returns:

            list: List of the names of required query parameters this method takes
        """
        return (
            [
                param_name
                for param_name in self.query_parameters
                if param_name in self.required_parameters
            ]
            if self.query_parameters
            else []
        )

    @property
    def optional_query_parameters(self) -> [str, str]:
        """
        Optional Query Parameters

        Returns:

            list: List of the names of optional query parameters this method takes
        """
        return (
            [
                param_name
                for param_name in self.query_parameters
                if param_name in self.optional_parameters
            ]
            if self.query_parameters
            else []
        )

    def __getitem__(self, key):
        """
        Examples:

            ::

                >>> self['description'] 
                
                "method description"

                >>> self['scopes']
                
                ['returns', 'scopes', 'required', 'by', 'this', 'method', 'in', 'a', 'list']

                >>> self['supportsMediaDownload']
                
                False

                >>> self['supportsMediaUpload']

                True

                >>> self['httpMethod']

                'GET'

        Hint:


            Using this method with ``scopes`` as an argument can be useful for incremental authorization. (Requesting scopes when needed. As opposed to requesting them at once)

            for more: https://developers.google.com/identity/protocols/OAuth2WebServer#incrementalAuth


        """
        return self._method_specs.get(key)

    def _validate(self, instance, schema, schema_name=None):
        return validate__(instance, schema, self._schemas, schema_name)

    @staticmethod
    def _rm_none_params(uri_params):
        for k, v in list(uri_params.items()):
            if v is None:
                del uri_params[k]
        return uri_params

    @_toggle2x_dashed_params
    def __call__(
        self,
        validate=None,
        data=None,
        json=None,
        upload_file=None,
        download_file=None,
        timeout=None,
        **uri_params,
    ) -> Request:
        """ 
        Builds a request from this method

        Note:

            * When passing ``datetime.datetime or datetime.date`` pass them in json format.
            
            * Aiogoogle won't do that as it would be a big hassle to iterate over every item in ``*uri_params``, ``json`` and ``data``
            to check if there's any datetime objects.

            * Fortunately Python makes it really easy to achieve that. 
            
            * Instead of passing say ``datetime.datetime.utcnow()``, pass: ``datetime.datetime.utcnow().jsonformat()``

        Note:

            * All ``None`` values are ommited before sending to Google apis, if you want to explicitly pass a JSON null then pass it as ``"null"`` not ``None`` 

        Arguments:

            validate (bool): Overrides :param: aiogoogle.Aiogoole.validate if not None
            
            json (dict): Json body
            
            data (any): Data body (Bytes, text, www-url-form-encoded and others)
            
            upload_file (str): full path of file to upload
            
            download_file (str): full path of file to download to
            
            timeout (str): total timeout for this request
            
            **uri_params (dict): path and query, required and optional parameters

        Returns:

            aiogoogle.models.Request: An unsent request object
        """
        # If collisions are found between the 'key' of **uri_params and explicit kwargs e.g. data, json etc., then
        # priority will be given to explicit kwargs. With that said, it's not likely there will be any.
        # If you want to double check if there are any collisions,
        # you can append the API name and version you're using to tests.globals.ALL_APIS (only if they don't exist, otherwise, you shouldn't worry about collisions)
        # Then, run the unit tests and monitor: tests.test_discovery_document.test_parameters_not_colliding_with_google_api__call__ for failure

        #
        # NOTE: Use '_' instead of '-' when passing uri parameters that have a '-' in their names
        #

        # Remove params that are None
        uri_params = self._rm_none_params(uri_params)

        # Assert timeout is int
        if timeout is not None:
            if not isinstance(timeout, int) or type(timeout) == bool:
                raise TypeError("Timeouts can only be ints or None")

        # Resolve validation status
        if not isinstance(validate, bool):
            validate = self._should_validate

        base_url = self._base_url

        # Build full url minus query & fragment
        url = self._build_url(
            base_url=base_url, uri_params=uri_params, validate=validate
        )

        # Filter out query parameters from all uri_params that were passed to this method
        passed_query_params = {
            param_name: param_info
            for param_name, param_info in uri_params.items()
            if param_name in self.query_parameters
        }

        # Ensure all required query parameteters were passed
        for param in self.required_query_parameters:
            if param not in passed_query_params:
                raise ValidationError(f'Missing query parameter: "{param}"')

        # Validate url query params
        if validate is True:
            if passed_query_params:
                for param_name, passed_param in passed_query_params.items():
                    self._validate(
                        passed_param,
                        self.parameters[param_name],
                        schema_name=param_name,
                    )

        # Join query params
        if passed_query_params:
            uri = url + "?" + urlencode(passed_query_params)
        else:
            uri = url

        # Pop uri_params consumed
        for param_name, _ in passed_query_params.items():
            del uri_params[param_name]

        # Warn if not all uri_params were consumed/popped
        if uri_params:  # should be empty by now
            # If there's room for addtionalProperties, validate and add them to the URI
            if validate:
                if self.parameters.get("additionalProperties"):
                    for _, v in uri_params.items():
                        self._validate(
                            v,
                            self.parameters["additionalProperties"],
                            schema_name="Additional Url Parameters",
                        )
                else:
                    raise ValidationError(
                        f"Invalid (extra) parameters: {uri_params} were passed"
                    )
            else:
                if not self.parameters.get("additionalProperties"):
                    warnings.warn(
                        "Parameters {} were found and they're probably of no use."
                        " Check if they're valid parameters".format(str(uri_params))
                    )
            if "?" not in uri:
                uri = uri + "?" + urlencode(uri_params)
            else:
                uri = uri + "&" + urlencode(uri_params)

        # Ensure only one param for http req body.
        if json and data:
            raise TypeError(
                "Pass either json or data for the body of the request, not both."
                '\nThis is similar to the "body" argument in google-python-client'
                "\nThis will validate agains the $request key in this method's "
                "specs"
            )  # This raises a TypeError instead of a ValidationError because
            # it will probably make your session raise an error if it passes.
            # Better raise it early on

        # Validate body
        if validate is True:
            body = (
                json if json is not None else data if data is not None else None
            )  # json or data or None
            if body is not None:
                self._validate_body(body)

        # Process download_file
        if download_file:
            if validate is True:
                if self.__getitem__("supportsMediaDownload") is not True:
                    raise ValidationError(
                        "download_file was provided while method doesn't support media download"
                    )
            media_download = MediaDownload(download_file)
        else:
            media_download = None

        # Process upload_file
        if upload_file:
            media_upload = self._build_upload_media(
                upload_file, uri, validate, fallback_url=url
            )
        else:
            media_upload = None

        return Request(
            method=self["httpMethod"],
            url=uri,
            batch_url=self._batch_url,
            data=data,
            json=json,
            timeout=timeout,
            media_download=media_download,
            media_upload=media_upload,
            callback=lambda res: self._validate_response(res, validate),
        )

    def _build_url(self, base_url, uri_params, validate):
        if self.path_parameters:
            # sort path params as sepcified in method_specs.parameterOrder
            sorted_required_path_params = (
                {}
            )  # Dict order is guaranteed (by insertion) as of Python 3.6
            for param_name in self["parameterOrder"]:
                try:
                    sorted_required_path_params[param_name] = uri_params.pop(param_name)
                except KeyError:
                    raise ValidationError(f"Missing URL path parameter: {param_name}")

            # Validate path params
            if validate is True:
                self._validate_url(sorted_required_path_params)

            # Build full path
            # replace named placeholders with empty ones. e.g. {param} --> {}
            # Why? Because some endpoints have different names in their url path placeholders than in their parameters
            # e.g. path: {"v1/{+resourceName}/connections"}. e.g. param name: resourceName NOT +resourceName
            self._method_specs["path"] = re.sub(
                r"\{(.*?)\}", r"{}", self._method_specs["path"]
            )
            return base_url + self["path"].format(
                *list(sorted_required_path_params.values())
            )
        else:
            return base_url + self["path"]

    def _build_upload_media(self, upload_file, qualified_url, validate, fallback_url):
        if self["supportsMediaUpload"] is not True:
            if validate is True:
                raise ValidationError(
                    "upload_file was provided while method doesn't support media upload"
                )
            else:
                # This will probably not work, but will return a mediaupload object anyway
                return MediaUpload(upload_file, upload_path=fallback_url)

        # If resumable, create resumable object
        resumable = (
            self._build_resumeable_media(upload_file, qualified_url)
            if _safe_getitem(
                self._method_specs, "mediaUpload", "protocols", "resumable"
            )
            else None
        )

        # Create MediaUpload object and pass it the resumable object we just created
        # simple_upload_path = _safe_getitem(self._method_specs, 'mediaUpload', 'protocols', 'simple', 'path')
        # media_upload_url_base = self._root_url[:-1] + simple_upload_path
        media_upload_url_base = self._root_url + "upload/" + self._service_path
        media_upload_url = qualified_url.replace(self._base_url, media_upload_url_base)
        max_size = self._convert_str_size_to_int(
            _safe_getitem(self._method_specs, "mediaUpload", "maxSize")
        )
        mime_range = _safe_getitem(self._method_specs, "mediaUpload", "accept")
        multipart = self["mediaUpload"]["protocols"]["simple"].get("multipart", True)

        # Return
        return MediaUpload(
            file_path=upload_file,
            upload_path=media_upload_url,
            max_size=max_size,
            mime_range=mime_range,
            multipart=multipart,
            resumable=resumable,
            validate=validate,
        )

    def _build_resumeable_media(self, upload_file, qualified_url):
        # resumable_upload_path = _safe_getitem(self._method_specs, 'mediaUpload', 'protocols', 'resumable', 'path')
        # resumable_url_base = self._root_url[:-1] + resumable_upload_path
        resumable_url_base = self._root_url + "resumable/upload/" + self._service_path
        resumable_url = qualified_url.replace(self._base_url, resumable_url_base)
        multipart = self["mediaUpload"]["protocols"]["resumable"].get("multipart", True)
        return ResumableUpload(
            upload_file, multipart=multipart, upload_path=resumable_url
        )

    @staticmethod
    def _convert_str_size_to_int(size):
        """Convert a string media size, such as 10GB or 3TB into an integer.

        Args:
            size: (str): e.g. such as 2MB or 7GB.

        Returns:
            The size as an integer value.
        """
        MEDIA_SIZE_BIT_SHIFTS = {"KB": 10, "MB": 20, "GB": 30, "TB": 40}
        if size is None:
            return None
        if len(size) < 2:
            return 0
        units = size[-2:].upper()
        bit_shift = MEDIA_SIZE_BIT_SHIFTS.get(units)
        if bit_shift is not None:
            return int(size[:-2]) << bit_shift
        else:
            return int(size)

    def _validate_url(self, sorted_required_path_params):
        for path_param_name, path_param_info in sorted_required_path_params.items():
            self._validate(
                instance=path_param_info,
                schema=self.parameters[path_param_name],
                schema_name=path_param_name,
            )

    def _validate_body(self, req):
        request_schema = self._method_specs.get("request")
        if request_schema is not None:
            schema_name = "Request Body"
            if "$ref" in request_schema:
                schema_name = request_schema["$ref"]
                request_schema = self._schemas[schema_name]
            self._validate(req, request_schema, schema_name=schema_name)
        else:
            raise ValidationError(
                "Request body should've been validated, but wasn't because the method doesn't accept a JSON body"
            )

    def _validate_response(self, res, validate):
        if validate is True and res is not None:
            schema_name = "Response_body"
            response_schema = self._method_specs.get("response")
            if response_schema is not None:
                if "$ref" in response_schema:
                    schema_name = response_schema["$ref"]
                    response_schema = self._schemas[schema_name]
            else:
                raise ValidationError(
                    "Response body should've been validated, but wasn't because a schema body wasn'nt found"
                )
            self._validate(res, response_schema, schema_name)
        return res

    def __contains__(self, item):
        return item in self.parameters

    def __str__(self):
        return self["id"] + " method @ " + self._base_url

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.required_parameters) if self.required_parameters else 0


class Resource:
    def __init__(
        self,
        name,
        resource_specs,
        global_parameters,
        schemas,
        root_url,
        service_path,
        batch_path,
        validate,
    ):
        self.name = name
        self._resource_specs = resource_specs
        self._global_parameters = global_parameters
        self._schemas = schemas
        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path
        self._validate = validate

    @property
    def methods_available(self) -> [str, str]:
        """
        Returns the names of the methods that this resource provides
        """
        return [k for k, v in self["methods"].items()] if self["methods"] else []

    @property
    def resources_available(self) -> [str, str]:
        """
        Returns the names of the nested resources in this resource
        """
        return [k for k, v in self["resources"].items()] if self["resources"] else []

    def __str__(self):
        return self.name + " resource @ " + self._root_url + self._service_path

    def __repr__(self):
        return self.__str__()

    def __call__(self):
        raise TypeError(
            "Only methods are callables, not resources."
            " e.g. api.resource.list() NOT api.resource().list()"
        )

    def __len__(self):
        return len(self.methods_available)

    def __contains__(self, item):
        return (item in self.methods_available) or (item in self.resources_available)

    def __getitem__(self, k):
        return self._resource_specs.get(k)

    def __getattr__(self, method_or_resource):
        """
        Returns either a method or a nested resource

        Arguments:

            method_or_resource: Name of the method or resource desired.

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Methods: A Resource or a Method


        Note:
            
            This method will first check in nested resources then will check in methods.

        Raises:

            AttributeError:
        """
        # 1. Search in nested resources
        if method_or_resource in self.resources_available:
            return Resource(
                name=method_or_resource,
                resource_specs=self["resources"][method_or_resource],
                global_parameters=self._global_parameters,
                schemas=self._schemas,
                root_url=self._root_url,
                service_path=self._service_path,
                batch_path=self._batch_path,
                validate=self._validate,
            )
        # 2. Search in methods
        elif method_or_resource in self.methods_available:
            return Method(
                name=method_or_resource,
                method_specs=self["methods"][method_or_resource],
                global_parameters=self._global_parameters,
                schemas=self._schemas,
                root_url=self._root_url,
                service_path=self._service_path,
                batch_path=self._batch_path,
                validate=self._validate,
            )
        else:
            raise AttributeError(
                f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: https://developers.google.com/ for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}"""
            )


class GoogleAPI:
    """
    Creetes a representation of Google API given a discovery document
    
    Arguments:

        discovery_document (dict): A discovery document

        validate (bool): Whther or not to validate user input again the schema defined in the discovery document
    """

    def __init__(self, discovery_document, validate=True):
        self.discovery_document = self._add_extra_params(discovery_document)
        self._validate = validate

    def _add_extra_params(self, discovery_document):
        """ Adds extra parameters that aren't mentioned in the discovery docuemnt """
        extra_params = {
            param: STACK_QUERY_PARAMETER_DEFAULT_VALUE
            for param in STACK_QUERY_PARAMETERS
        }
        if discovery_document.get("parameters"):
            discovery_document["parameters"] = {
                **discovery_document["parameters"],
                **extra_params,
            }
        else:
            discovery_document["parameters"] = extra_params
        return discovery_document

    @property
    def methods_available(self) -> [str, str]:
        """
        Returns names of the methods provided by this resource
        """
        return [k for k, v in self["methods"].items()] if self["methods"] else []

    @property
    def resources_available(self) -> [str, str]:
        """
        Returns names of the resources in a given API if any
        """
        return [k for k, v in self["resources"].items()] if self["resources"] else []

    def __getattr__(self, method_or_resource) -> Resource:
        """
        Returns resources from an API

        Note:
            
            This method will first check in resources then will check in methods.

        Arguments:

            method_or_resource (str): name of the top level method or resource

        Example:

            ::

                >>> google_service = GoogleAPI(google_service_discovery_doc)
                >>> google_service.user

                user resource @ google_service.com/api/v1/

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Method: A Resource or a Method

        Raises:

            AttributeError:
        """
        if method_or_resource in self.resources_available:
            return Resource(
                name=method_or_resource,
                resource_specs=self["resources"][method_or_resource],
                global_parameters=self["parameters"],
                schemas=self["schemas"]
                or {},  # josnschema validator will fail if schemas isn't a dict
                root_url=self["rootUrl"],
                service_path=self["servicePath"],
                batch_path=self["batchPath"],
                validate=self._validate,
            )
        elif method_or_resource in self.methods_available:
            return Method(
                name=method_or_resource,
                method_specs=self["methods"][method_or_resource],
                global_parameters=self["parameters"],
                schemas=self["schemas"]
                or {},  # josnschema validator will fail if schemas isn't a dict
                root_url=self["rootUrl"],
                service_path=self["servicePath"],
                batch_path=self["batchPath"],
                validate=self._validate,
            )
        else:
            documentation_link = (
                self.discovery_document.get("documentationLink")
                or "https://developers.google.com/"
            )
            raise AttributeError(
                f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: {documentation_link} for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}"""
            )

    def __getitem__(self, k):
        """
        Returns items from the discovery document

        Example:

            ::

                >>> google_service = GoogleAPI(google_books_discovery_doc)
                >>> google_service['name']

                'books'

                >>> google_service['id']

                'books:v1'

                >>> google_service['version']

                'v1'

                >>> google_service['documentationLink']

                'https://developers.google.com/books/docs/v1/getting_started'                
                
                >>> google_service['oauth2']['scopes']

                https://www.googleapis.com/auth/books: {
                    
                    description: "Manage your books"
                
                }

        Returns:

            dict: Discovery Document Item
        """
        return self.discovery_document.get(k)

    def __contains__(self, item):
        return (item in self.resources_available) or (item in self.methods_available)

    def __repr__(self):
        labels = f'\nLabels:\n{self["labels"]}' if self["labels"] is not None else ""
        return (
            self.discovery_document["name"]
            + "-"
            + self.discovery_document["version"]
            + " API @ "
            + self["rootUrl"]
            + self["servicePath"]
            + labels
        )

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self.resources_available) + len(self.methods_available)

    def __call__(self):
        raise TypeError(
            "Only methods are callables, not resources."
            " e.g. client.resources.user.list() NOT client.resources().user.list()"
        )



def _call_callback(request, response):
    if request.callback is not None:
        if (
            response.status_code is not None
            and response.status_code >= 200
            and response.status_code < 300
        ):
            if response.json is not None:
                response.json = request.callback(response.content)
            elif response.data is not None:
                response.data = request.callback(response.content)
    return response



import setuptools

import os
from os import listdir
from os.path import isfile, join

mypath = os.path.dirname(os.path.abspath(__file__))
print([f for f in listdir(mypath) if isfile(join(mypath, f))])


with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

with open("test_requirements.txt", "r") as f:
    test_requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiogoogle",
    version="0.1.11",
    author="Omar Ryhan",
    author_email="omarryhan@gmail.com",
    license="GNU",
    description="Async Discovery API Client + Authentication",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    tests_require=test_requirements,
    url="https://github.com/omarryhan/aiogoogle",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    extras_require={"curio_asks": ["asks", "curio"], "trio_asks": ["asks", "trio"]},
)



"""
.. note::

    * These are the default auth managers. They won't perform any file io.

    * If you want auth managers with file io capabilities, then you'll have to implement AbstractAuthManager's interface or inherent from any of this module's managers.
    
    * In most cases you won't need to implement new managers for file io, as by design, credentials are an instance of dict and will only contain json types (str, number, array, JSONSCHEMA datetime and ISO8601 datetime, etc) to make it easily serializable.
"""

__all__ = ["ApiKeyManager", "Oauth2Manager", "OpenIdConnectManager"]

from urllib import parse
import datetime

try:
    import ujson as json
except:
    import json
from google.auth import jwt

from .abc import (
    AbstractOAuth2Manager,
    AbstractAPIKeyManager,
    AbstractOpenIdConnectManager,
)
from .creds import UserCreds, ClientCreds
from .data import OAUTH2_V2_DISCVOCERY_DOC, WELLKNOWN_OPENID_CONFIGS
from ..excs import HTTPError, AuthError
from ..models import Request
from ..resource import GoogleAPI
from ..sessions.aiohttp_session import AiohttpSession


# ID token contents: https://openid.net/specs/openid-connect-core-1_0.html#IDToken
# JWK reference: https://tools.ietf.org/html/rfc7517
# Google OpenID discovery doc reference: https://developers.google.com/identity/protocols/OpenIDConnect#discovery

# discovery docs
OPENID_CONFIGS_DISCOVERY_DOC_URL = (
    "https://accounts.google.com/.well-known/openid-configuration"
)
OAUTH2_DISCOVERY_DOCUMENT_URL = (
    "https://www.googleapis.com/discovery/v1/apis/oauth2/v2/rest"
)

# Response types
AUTH_CODE_RESPONSE_TYPE = "code"  # token for implicit flow and and openid for OpenID
HYBRID_RESPONSE_TYPE = "code id_token"

# Grant types
AUTH_CODE_GRANT_TYPE = "authorization_code"
REFRESH_GRANT_TYPE = "refresh_token"
JWT_GRANT_TYPE = (
    "urn:ietf:params:oauth:grant-type:jwt-bearer"
)  # https://tools.ietf.org/html/rfc7523#section-4  OAuth JWT Assertion Profiles

# Other types
URLENCODED_CONTENT_TYPE = "application/x-www-form-urlencoded"
# _Installed Application Authorization Flow:
#  https://developers.google.com/api-client-library/python/auth/installed-app
OOB_REDIRECT_URI = "urn:ietf:wg:oauth:2.0:oob"

DEFAULT_ISS = "https://accounts.google.com"

## The URL that provides public certificates for verifying ID tokens issued
## by Google's OAuth 2.0 authorization server.
GOOGLE_OAUTH2_CERTS_URL = "https://www.googleapis.com/oauth2/v1/certs"
# GOOGLE_OAUTH2_CERTS_URL_V3 = 'https://www.googleapis.com/oauth2/v3/certs'

# Used here: https://developers.google.com/identity/protocols/OAuth2WebServer#exchange-authorization-code
# AUTH_URI = 'https://accounts.google.com/o/oauth2/v2/auth'
# REVOKE_URI = 'https://accounts.google.com/o/oauth2/revoke'
# TOKEN_URI = REFRESH_URI = 'https://www.googleapis.com/oauth2/v4/token'  # Not included in OAuth2 v2 discovery doc
TOKEN_INFO_URI = "https://www.googleapis.com/oauth2/v4/tokeninfo"

## Used here: https://github.com/googleapis/oauth2client/blob/master/oauth2client/__init__.py
## Google oauth2client is deprecated in favor of google-oauth-lib & google-auth-library-python
## These aren't used in the code below in favor of the urls above
# GOOGLE_AUTH_URI = 'https://accounts.google.com/o/oauth2/v2/auth'
# GOOGLE_DEVICE_URI = 'https://oauth2.googleapis.com/device/code'
# GOOGLE_REVOKE_URI = 'https://oauth2.googleapis.com/revoke'
# GOOGLE_TOKEN_URI = 'https://oauth2.googleapis.com/token'
# GOOGLE_TOKEN_INFO_URI = 'https://oauth2.googleapis.com/tokeninfo'


class ApiKeyManager(AbstractAPIKeyManager):
    def __init__(self, api_key=None):
        self.key = api_key

    @staticmethod
    def authorize(request, key=None) -> Request:
        key = key or self.key
        if "key=" in request.url:
            return request
        else:
            url = request.url
            if "?" not in url:
                if url.endswith("/"):
                    url = url[:-1]
                url += "?"
            else:
                url += "&"
            url += f"key={key}"
            request.url = url
            return request


class Oauth2Manager(AbstractOAuth2Manager):
    def __init__(self, session_factory=AiohttpSession, verify=True, client_creds=None):
        self.oauth2_api = GoogleAPI(OAUTH2_V2_DISCVOCERY_DOC)
        self.openid_configs = WELLKNOWN_OPENID_CONFIGS
        self.session_factory = session_factory
        self.active_session = None
        self.verify = verify
        self.client_creds = client_creds

    def __getitem__(self, key):
        """
        Gets Google's openID configs

        Example:

            * response_types_supported
            
            * scopes_supported
            
            * claims_supported
        """
        try:
            return self.openid_configs[key]
        except KeyError:
            raise

    async def __aenter__(self):
        self.active_session = await self.session_factory().__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.active_session.__aexit__(*args, **kwargs)
        self.active_session = None

    async def _send_request(self, req):
        if self.active_session is None:
            async with self.session_factory() as sess:
                res = await sess.send(req)
        else:
            res = await self.active_session.send(req)
        return res

    async def _refresh_openid_configs(self):
        """
        Downloads fresh openid discovery document and sets it to the current manager.

        OpenID configs are used for both OAuth2 manager and OpenID connect manager.

        Unless this test is failing:

        aiogoogle.tests.integ_online.latest_test_latest_openid_configs(), You shouldn't really need to use this. 
        """
        req = Request("GET", url=OPENID_CONFIGS_DISCOVERY_DOC_URL)
        self.openid_configs = await self._send_request(req)

    @staticmethod
    def authorize(request: Request, user_creds: dict) -> Request:
        if request.headers is None:
            request.headers = {}
        request.headers["Authorization"] = f'Bearer {user_creds["access_token"]}'
        return request

    @staticmethod
    def is_ready(client_creds=None):
        """
        Checks passed ``client_creds`` whether or not the client has enough information to perform OAuth2 Authorization code flow

        Arguments:

            client_creds(aiogoogle.auth.creds.ClientCreds): Client credentials object

        Returns:

            bool: 
        """
        client_creds = client_creds or self.client_creds
        if (
            client_creds.get("client_id")
            and client_creds.get("client_secret")
            and client_creds.get("scopes")
            and isinstance(client_creds["scopes"], (list, tuple))
            and client_creds.get("redirect_uri")
        ):
            return True
        return False

    def authorization_url(
        self,
        client_creds=None,
        state=None,
        access_type=None,
        include_granted_scopes=None,
        login_hint=None,
        prompt=None,
        response_type=AUTH_CODE_RESPONSE_TYPE,
        scopes=None,
    ) -> (str):
        client_creds = client_creds or self.client_creds
        scopes = scopes or client_creds["scopes"]
        scopes = " ".join(client_creds["scopes"])
        uri = (
            self["authorization_endpoint"]
            + f'?redirect_uri={client_creds["redirect_uri"]}&scope={scopes}&'
        )
        for param_name, param in {
            "client_id": client_creds["client_id"],
            "response_type": response_type,
            "state": state,
            "access_type": access_type,
            "include_granted_scopes": json.dumps(include_granted_scopes),
            "login_hint": login_hint,
            "prompt": prompt,
        }.items():
            if param is not None:
                uri += "&"
                uri += parse.urlencode({param_name: param})
        return uri

    async def build_user_creds(
        self, grant, client_creds=None, grant_type=AUTH_CODE_GRANT_TYPE
    ) -> dict:
        client_creds = client_creds or self.client_creds
        request = self._build_user_creds_req(grant, client_creds, grant_type)
        json_res = await self._send_request(request)
        return self._build_user_creds_from_res(json_res)

    def _build_user_creds_req(self, grant, client_creds, grant_type) -> Request:
        data = dict(
            code=grant,
            client_id=client_creds["client_id"],
            client_secret=client_creds["client_secret"],
            redirect_uri=client_creds["redirect_uri"],
            grant_type=grant_type,
        )
        headers = {"content-type": URLENCODED_CONTENT_TYPE}
        method = "POST"
        url = self["token_endpoint"]
        return Request(method, url, headers, data=data)

    def _build_user_creds_from_res(self, json_res):
        scopes = json_res.pop("scope").split(" ")
        user_creds = UserCreds(**json_res, scopes=scopes)
        # Idk why, but sometimes google returns these json params empty
        user_creds["token_uri"] = (
            self["token_endpoint"] if user_creds.get("token_uri") is None else None
        )
        user_creds["token_info_uri"] = (
            TOKEN_INFO_URI if not user_creds.get("token_info_uri") else None
        )
        user_creds["revoke_uri"] = (
            self["revocation_endpoint"] if not user_creds.get("revoke_uri") else None
        )
        user_creds["expires_at"] = self._get_expires_at(user_creds["expires_in"])
        return user_creds

    @staticmethod
    def _get_expires_at(expires_in):
        expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=expires_in)
        # account for clock skew
        expires_at -= datetime.timedelta(seconds=120)
        return expires_at.isoformat()

    @staticmethod
    def authorized_for_method(method, user_creds) -> bool:
        """
        Checks if oauth2 user_creds dict has sufficient scopes for a method call.
        
        .. note:: 
        
            This method doesn't check whether creds are refreshed or valid.

        e.g.

            **Correct:**

            .. code-block:: python3
        
                is_authorized = authorized_for_method(youtube.video.list)
            
            **NOT correct:**

            .. code-block:: python3

                is_authorized = authorized_for_method(youtube.video.list())

            **AND NOT correct:**

            .. code-block:: python3

                is_authorized = authorized_for_method(youtube.videos)


        Arguments:

            method (aiogoogle.resource.Method): Method to be checked

            user_credentials (aiogoogle.auth.creds.UserCreds): User Credentials with scopes item

        Returns:

            bool:

        """
        method_scopes = method["scopes"] or []
        if not method_scopes:
            return True
        if not isinstance(user_creds["scopes"], (list, set, tuple)):
            raise TypeError("Scopes should be an instance of list, set or tuple")
        if set(method_scopes).issubset(set(user_creds["scopes"])):
            return True
        else:
            return False

    async def get_token_info(self, user_creds):
        """
        Gets token info given an access token

        Arguments:

            user_creds (aiogoogle.creds.UserCreds): UserCreds instance with an access token

        Returns:

            dict: Info about the token

        """
        req = self.oauth2_api.tokeninfo(access_token=user_creds.get("access_token"))
        return await self._send_request(req)

    async def get_me_info(self, user_creds):
        """
        Gets information of a user given his access token. User must also be the client.
        (Not sure whether or not that's the main purpose of this endpoint and how it differs from get_user_info.
        If someone can confirm/deny the description above, please edit (or remove) this message and make a pull request)

        Arguments:

            user_creds (aiogoogle.creds.UserCreds): UserCreds instance with an access token

        Returns:

            dict: Info about the user

        Raises:

            aiogoogle.excs.HTTPError:
         """
        req = self.oauth2_api.userinfo.v2.me.get()
        authorized_req = self.authorize(req, user_creds)
        return await self._send_request(authorized_req)

    @staticmethod
    def is_expired(creds) -> bool:
        expires_at = creds["expires_at"]
        if not isinstance(expires_at, datetime.datetime):
            expires_at = datetime.datetime.fromisoformat(expires_at)
        if datetime.datetime.utcnow() >= expires_at:
            return True
        else:
            return False

    async def refresh(self, user_creds, client_creds=None):
        client_creds = client_creds or self.client_creds
        request = self._build_refresh_request(user_creds, client_creds)
        json_res = await self._send_request(request)
        return self._build_user_creds_from_res(json_res)

    def _build_refresh_request(self, user_creds, client_creds):
        data = dict(
            grant_type=REFRESH_GRANT_TYPE,
            client_id=client_creds["client_id"],
            client_secret=(client_creds["client_secret"]),
            refresh_token=(user_creds["refresh_token"]),
        )
        method = "POST"
        headers = {"content-type": URLENCODED_CONTENT_TYPE}
        return Request(method, self["token_endpoint"], headers, data=data)

    def _build_revoke_request(self, user_creds):
        return Request(
            method="POST",
            headers={"content-type": URLENCODED_CONTENT_TYPE},
            url=self["revocation_endpoint"],
            data=dict(token=user_creds["access_token"]),
        )

    async def revoke(self, user_creds):
        request = self._build_revoke_request(user_creds)
        return await self._send_request(request)


class OpenIdConnectManager(Oauth2Manager, AbstractOpenIdConnectManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def get_user_info(self, user_creds):
        req = self.oauth2_api.userinfo.get()
        authorized_req = self.authorize(req, user_creds)
        return await self._send_request(authorized_req)

    async def get_token_info_jwt(self, user_creds):
        """ get token info using id_token_jwt instead of access_token ``self.get_token_info``
        
        Arguments:
        
            user_creds (aiogoogle.auth.creds.UserCreds): user_creds with id_token_jwt item
            
        Returns:
        
            dict: Information about the token
        """
        req = self.oauth2_api.tokeninfo(id_token=user_creds.get("id_token_jwt"))
        return await self._send_request(req)

    async def _get_openid_certs(self):
        request = Request(
            method="GET",
            url=GOOGLE_OAUTH2_CERTS_URL,
            # url=self['jwks_uri'],  # which is: https://www.googleapis.com/oauth2/v3/certs isn't compatible with google.auth. Falling back to v1
        )
        return await self._send_request(request)

    def authorization_url(
        self,
        client_creds=None,
        nonce=None,
        state=None,
        prompt=None,
        display=None,
        login_hint=None,
        access_type=None,
        include_granted_scopes=None,
        openid_realm=None,
        hd=None,
        response_type=AUTH_CODE_RESPONSE_TYPE,
        scopes=None,
    ):
        client_creds = client_creds or self.client_creds
        if nonce is None:
            raise TypeError("Nonce is required")
        scopes = scopes or client_creds["scopes"]
        scopes = " ".join(scopes)
        uri = (
            self["authorization_endpoint"]
            + f'?redirect_uri={client_creds["redirect_uri"]}&scope={scopes}&'
        )
        for param_name, param in {
            "client_id": client_creds["client_id"],
            "nonce": nonce,
            "display": display,
            "openid.realm": openid_realm,
            "hd": hd,
            "response_type": response_type,
            "state": state,
            "access_type": access_type,
            "include_granted_scopes": json.dumps(include_granted_scopes),
            "login_hint": login_hint,
            "prompt": prompt,
        }.items():
            if param is not None:
                uri += "&"
                uri += parse.urlencode({param_name: param})
        return uri

    async def build_user_creds(
        self,
        grant,
        client_creds=None,
        grant_type=AUTH_CODE_GRANT_TYPE,
        nonce=None,
        hd=None,
        verify=True,
    ):
        client_creds = client_creds or self.client_creds
        user_creds = await super().build_user_creds(
            grant, client_creds, grant_type=grant_type
        )
        user_creds["id_token_jwt"] = user_creds["id_token"]
        if verify is False:
            user_creds["id_token"] = jwt.decode(
                user_creds["id_token_jwt"], verify=False
            )
        else:
            user_creds["id_token"] = await self.decode_and_validate(
                user_creds["id_token_jwt"], client_creds["client_id"], nonce, hd
            )
        return user_creds

    async def decode_and_validate(
        self, id_token_jwt, client_id=None, nonce=None, hd=None
    ):
        certs = await self._get_openid_certs()  # refreshed once a day-ish
        # Verify ID token is signed by google
        try:
            id_token = jwt.decode(id_token_jwt, certs=certs, verify=True)
        except ValueError as e:
            raise AuthError(e)
        # Verify iss (The Issuer Identifier for the Issuer of the response) is https://accounts.google.com
        if id_token["iss"] != DEFAULT_ISS:
            raise AuthError(
                f"Invalid issuer, got: {id_token['iss']}, expected: {DEFAULT_ISS}"
            )
        if nonce is not None:
            if nonce != id_token["nonce"]:
                raise AuthError("Provided nonce does not match the encoded nonce")
        if hd is not None:
            if hd != id_token["hd"]:
                raise AuthError(
                    f"Hosted domains do not match, got: {id_token['hd']}, expected: {hd}"
                )
        # verify expiry 'exp' (google.jwt handles that)
        # verify audience
        if client_id is not None:
            if id_token["aud"] != client_id:
                raise AuthError(
                    f"Invalid audience. Got: {id_token['aud']} expected: {client_id}"
                )
        return id_token

    async def build_user_creds_jwt_grant(self, assertion, token_uri):
        """
        Implements the JWT Profile for OAuth 2.0 Authorization Grants.

        Args:

            assertion (str):

                * A single id_token_jwt

            token_uri (str):

                * Token URI of your authorization server

        Returns:
            
            aiogoogle.auth.creds.UserCreds:

        Raises:
            
            aiogoogle.excs.AuthError: 

        .. _rfc7523 section 4: https://tools.ietf.org/html/rfc7523#section-4 (Section 2.1 for an example: https://tools.ietf.org/html/rfc7523#section-2.1)
        """
        data = {"assertion": assertion, "grant_type": JWT_GRANT_TYPE}
        req = Request(method="POST", url=token_uri, data=data)
        json_res = await self._send_request(req)
        user_creds = self._build_user_creds_from_res(json_res)
        if user_creds.get("id_token"):  # Google specific (Not RFC compliant)
            user_creds["id_token"] = jwt.decode(
                user_creds["id_token_jwt"], verify=False
            )
        return user_creds

    async def build_user_creds_jwt_auth(self, assertion, token_uri):
        """https://tools.ietf.org/html/rfc7523#section-2.2 """
        raise NotImplementedError
